@@ -9,9 +9,10 @@ from pint import UnitRegistry
 from pint.unit import UnitsContainer
 from pint.util import ParserHelper
 
-from pint.compat import np, unittest, long_type
+from pint.compat import np, long_type
 from pint.errors import UndefinedUnitError
 from pint.testsuite import QuantityTestCase, helpers
+from pint.testsuite.compat import unittest
 
 
 class TestIssues(QuantityTestCase):
@@ -233,7 +234,7 @@ class TestIssues(QuantityTestCase):
 
         for func in (ureg.get_name, ureg.parse_expression):
             val = func('meter')
-            self.assertRaises(ValueError, func, 'METER')
+            self.assertRaises(AttributeError, func, 'METER')
             self.assertEqual(val, func('METER', False))
 
     def test_issue104(self):
@@ -499,7 +500,8 @@ class TestIssuesNP(QuantityTestCase):
         self.assertEqual(np.float16(a/b), 1000.)
         self.assertEqual(np.float32(a/b), 1000.)
         self.assertEqual(np.float64(a/b), 1000.)
-        self.assertEqual(np.float128(a/b), 1000.)
+        if "float128" in dir(np):
+            self.assertEqual(np.float128(a/b), 1000.)
 
     def test_issue252(self):
         ur = UnitRegistry()
@@ -519,3 +521,36 @@ class TestIssuesNP(QuantityTestCase):
         self.assertEqual(q1.units, self.ureg.dimensionless)
         q2 = self.ureg('1 dimensionless')
         self.assertEqual(q1, q2)
+
+    def test_issue354_356_370(self):
+        q = 1 * self.ureg.second / self.ureg.millisecond
+        self.assertEqual('{0:~}'.format(1 * self.ureg.second / self.ureg.millisecond),
+                         '1.0 s / ms')
+        self.assertEqual("{0:~}".format(1 * self.ureg.count),
+                         '1 count')
+        self.assertEqual('{0:~}'.format(1 * self.ureg('MiB')),
+                         '1 MiB')
+
+    def test_issue482(self):
+        q = self.ureg.Quantity(1, self.ureg.dimensionless)
+        qe = np.exp(q)
+        self.assertIsInstance(qe, self.ureg.Quantity)
+
+    def test_issue468(self):
+        ureg = UnitRegistry()
+
+        @ureg.wraps(('kg'), 'meter')
+        def f(x):
+            return x
+
+        x = ureg.Quantity(1., 'meter')
+        y = f(x)
+        z = x * y
+        self.assertEquals(z, ureg.Quantity(1., 'meter * kilogram'))
+
+    def test_issue483(self):
+        ureg = self.ureg
+        a = np.asarray([1, 2, 3])
+        q = [1, 2, 3] * ureg.dimensionless
+        p = (q ** q).m
+        np.testing.assert_array_equal(p, a ** a)
\ No newline at end of file
