@@ -192,6 +192,8 @@ def solve_dependencies(dependencies):
         # and keys without value (items without dep)
         t.update(k for k, v in d.items() if not v)
         # can be done right away
+        if not t:
+            raise ValueError('Cyclic dependencies exist among these items: {}'.format(', '.join(repr(x) for x in d.items())))
         r.append(t)
         # and cleaned up
         d = dict(((k, v - t) for k, v in d.items() if v))
@@ -325,6 +327,9 @@ class UnitsContainer(Mapping):
     def __format__(self, spec):
         return format_unit(self, spec)
 
+    def format_babel(self, spec, **kwspec):
+        return format_unit(self, spec, **kwspec)
+
     def __copy__(self):
         return UnitsContainer(self._d)
 
@@ -410,14 +415,15 @@ class ParserHelper(UnitsContainer):
     
     @classmethod
     def eval_token(cls, token, use_decimal=False):
-        token_type = token[0]
-        token_text = token[1]
+        token_type = token.type
+        token_text = token.string
         if token_type == NUMBER:
-            if '.' in token_text or 'e' in token_text:
+            try:
+                return int(token_text)
+            except ValueError:
                 if use_decimal:
                     return Decimal(token_text)
                 return float(token_text)
-            return int(token_text)
         elif token_type == NAME:
             return ParserHelper.from_word(token_text)
         else:
@@ -548,7 +554,8 @@ class ParserHelper(UnitsContainer):
 
 
 #: List of regex substitution pairs.
-_subs_re = [(r"([\w\.\-\+\*\\\^])\s+", r"\1 "), # merge multiple spaces
+_subs_re = [('\N{DEGREE SIGN}', " degree"),
+            (r"([\w\.\-\+\*\\\^])\s+", r"\1 "), # merge multiple spaces
             (r"({0}) squared", r"\1**2"),  # Handle square and cube
             (r"({0}) cubed", r"\1**3"),
             (r"cubic ({0})", r"\1**3"),
@@ -634,14 +641,14 @@ def to_units_container(unit_like, registry=None):
 
 def infer_base_unit(q):
     """Return UnitsContainer of q with all prefixes stripped."""
-    d = {}
+    d = udict()
     parse = q._REGISTRY.parse_unit_name
     for unit_name, power in q._units.items():
         completely_parsed_unit = list(parse(unit_name))[-1]
 
         _, base_unit, __ = completely_parsed_unit
-        d[base_unit] = power
-    return UnitsContainer(d)
+        d[base_unit] += power
+    return UnitsContainer(dict((k, v) for k, v in d.items() if v != 0))  # remove values that resulted in a power of 0
 
 
 def fix_str_conversions(cls):
@@ -689,7 +696,7 @@ class SourceIterator(object):
         line = ''
         while not line or line.startswith('#'):
             lineno, line = next(self.internal)
-            line = line.strip()
+            line = line.split('#', 1)[0].strip()
 
         self.last = lineno, line
         return lineno, line
